c-----------------------------------------------------------------------
c ----------------  Potential-dependent subroutines --------------------
c-----------------------------------------------------------------------


c ------general routines, which redirect for the specific potential-------

      subroutine potset(ndim,iicase)
      implicit real*8(a-h,o-z)
      common/potcom/icase

c  allowed values for icase:
c    1 = model potential I from A.F. Voter, J. Chem. Phys. 106, 4665 (1997).
c    2 = model potential II from A.F. Voter, J. Chem. Phys. 106, 4665 (1997).
c    3 = Frenkel-Kontorova model

      icase=iicase

      if(icase.le.2) then
        call potset_jcp(ndim, icase)
      else if(icase.eq.3) then
        call potset_fk(ndim)
      else
        write(6,*) 'undefined potential case:',icase
        stop 'potset: undefined potential case'
      end if
    
      return
      end


      subroutine v(x,n, energy)
      implicit real*8(a-h,o-z)
      dimension x(n)
      common/potcom/icase

      if(icase.le.2) then
        call v_jcp(x,n, energy)
      else if(icase.eq.3) then
        call v_fk(x,n, energy)
      else
        write(6,*) 'undefined potential case:',icase
        stop 'v: undefined potential case'
      end if
    
      return
      end


      subroutine vp(x,n, dvdx)
      implicit real*8(a-h,o-z)
      dimension x(n),dvdx(n)
      common/potcom/icase

      if(icase.le.2) then
        call vp_jcp(x,n, dvdx)
      else if(icase.eq.3) then
        call vp_fk(x,n, dvdx)
      else
        write(6,*) 'undefined potential case:',icase
        stop 'vp: undefined potential case'
      end if
    
      return
      end


      integer function kstate(x,n)
      implicit real*8(a-h,o-z)
      common/potcom/icase
      dimension x(n)

      if(icase.le.2) then
        kstate=kstate_jcp(x,n)
      else if(icase.eq.3) then
        kstate=kstate_fk(x,n)
      else
        write(6,*) 'undefined potential case:',icase
        stop 'kstate: undefined potential case'
      end if

      return
      end
     

      real*8 function vbias_flat(xx, n, energy, dvdx, vflat)
c flat bias potential proposed in:
c        M. M. Steiner, P.-A. Genilloud, and J. W. Wilkins,
c        Phys. Rev. B   57, 10236 (1998).
c take unbiased position and energy, return bias potential and 
c   derivative of V + V_bias
      implicit real*8(a-h,o-z)
      dimension xx(n),dvdx(n)

      if(energy.ge.vflat) then    ! no bias, regular derivatives
         vbias_flat = 0.0d0
         call vp(xx,n, dvdx)
      else                      ! flat potential, zero derivatives
         vbias_flat = vflat-energy
         call dmzer(dvdx,n)
      end if

      return
      end

c----------------------------------------------------------
c----------- routines for specific potentials -------------
c----------------------------------------------------------

c--------routines for potential from first hyperdynamics paper---------

      subroutine potset_jcp(ndim, icase)
c  Set the parameters that define the potential.
      implicit real*8(a-h,o-z)
      common/potcom1/d1,d2,d3,d4,d5,pi,twopi,twopi2

      if (icase .eq. 1) then
c  model potential I from A.F. Voter, J. Chem. Phys. 106, 4665 (1997).
         d1=4.0d0
         d2=1.0d0
         d3=0.0d0
         d4=1.0d0
      else if (icase .eq. 2) then
c model potential II
         d1 = 4.d0
         d2 = 1.d0
         d3 = -0.75d0
         d4 = 3.d0
      end if
      d5=0.d0                  ! for biased diffusion (not in the JCP paper)

      write(6,*) '-------------------------------------------------'
      write(6,*) 'parameters for model potential from Voter JCP 1997:'
      write(6,10) d1, d2, d3, d4
 10   format('  d1 =',f6.3, '  d2 =',f6.3, '  d3 =',f6.3, '  d4 =',f6.3)
      if(d5.ne.0.0d0) then
         write(6,12) d5
         write(6,*) '***note*** crossing counts may have little meaning'
         write(6,*) '           if d5 is turned on, moving saddles'
      end if
 12   format('     d5 =',f6.3,' (biased diffusion)')
      write(6,*) '-------------------------------------------------'


c some constants
      pi=2.0d0*acos(0.0d0)
      twopi=2.0d0*pi
      twopi2=twopi**2
  
      return
      end


      subroutine mep_jcp(n, zmax, a, vflat)
      implicit real*8(a-h,o-z)
      dimension xx(n), dvdx(n)
      common/potcom1/d1,d2,d3,d4,d5,pi,twopi,twopi2

c compute minimum energy pathway, V(x, ymin), dump to fort.15
      xmax = 2.d0
      if (d3.ne.0.d0) xmax = 2.d0 * d4    ! 0 -> xmax = two periods
      xx(1) = 0.d0
      do while (xx(1) .le. xmax)

c quick scan for global minima of original and biased potential slices
         vmin = 1.d5
         vbiasmin = 1.d5
         do i = 0, 100
c for n > 2, restrict search to symmetric points 
            do j = 2, n
               xx(j) = -0.5d0 + i * 0.01d0
            end do
            call v_jcp(xx, n, vnow)
            if (zmax .lt. 0.d0) then
               vbiasnow = vnow + vbias_flat(xx, n, vnow, dvdx, vflat)
            else
               vbiasnow = vnow + vbias_jcp(xx, n, vnow, dvdx, zmax, a)
            end if

            if (vnow .lt. vmin) then
               vmin = vnow
               ymin = xx(2)
            end if
            if (vbiasnow .lt. vbiasmin) then
               vbiasmin = vbiasnow
               ybiasmin = xx(2)
            end if
         end do

c bisection to zoom in on minima
         dy = 0.005d0
         do i = 1, 10
            do j = 2, n
               xx(j) = ymin + dy
            end do
            call v_jcp(xx, n, vmove)
            if (vmove .lt. vmin) then
               vmin = vmove
               ymin = xx(2)
            else
               do j = 2, n
                  xx(j) = ymin - dy
               end do
               call v_jcp(xx, n, vmove)
               if (vmove .lt. vmin) then
                  vmin = vmove
                  ymin = xx(2)
               end if
            end if

            do j = 2, n
               xx(j) = ybiasmin + dy
            end do
            call v_jcp(xx, n, vmove)
            if (zmax .lt. 0.d0) then
               vmove = vmove + vbias_flat(xx, n, vmove, dvdx, vflat)
            else
               vmove = vmove + vbias_jcp(xx, n, vmove, dvdx, zmax, a)
            end if
            if (vmove .lt. vbiasmin) then
               vbiasmin = vmove
               ybiasmin = xx(2)
            else
               do j = 2, n
                  xx(j) = ybiasmin - dy
               end do
               call v_jcp(xx, n, vmove)
               if (zmax .lt. 0.d0) then
                  vmove = vmove + vbias_flat(xx, n, vmove, dvdx, vflat)
               else
                  vmove = vmove + vbias_jcp(xx, n, vmove, dvdx, zmax, a)
               end if
               if (vmove .lt. vbiasmin) then
                  vbiasmin = vmove
                  ybiasmin = xx(2)
               end if
            end if

            dy = 0.5d0 * dy
         end do

         write(15,*) xx(1), vmin, vbiasmin
         xx(1) = xx(1) + 0.02d0
      end do
      write(6,*) '------ V(x, ymin) paths written to unit 15 ------'
      write(6,*) '-------------------------------------------------'

      return
      end


      subroutine v_jcp(xx,n, energy)
c "cos + parabola"  - 2D function of x and y, periodic in x, as in JCP paper
c
c-------------------------------------------------------------
c  note:  there is a typo in the JCP paper:  
c         the 2*pi factor in d2 term should be (2*pi)**2
c-------------------------------------------------------------

      implicit real*8(a-h,o-z)
      dimension xx(n)
      common/potcom1/d1,d2,d3,d4,d5,pi,twopi,twopi2
     
ccc      if(n.ne.2) stop 'need n=2 in subroutine v'
      
      x=xx(1)
      y=xx(2)

      energy = cos(twopi*x)*(1.0d0 + d1*y) + d2*0.5d0*twopi2*y**2
     x         + d3*cos(twopi*x/d4) + d5*x

      if(n.gt.2) then
      do i=3,n
         energy=energy+cos(twopi*x)*(d1*xx(i))+d2*0.5d0*twopi2*xx(i)**2
      end do
      end if

c tcg - potential shift to put minima at v = 0
      energy = energy + 1.d0 + 2.d0 * (n-1.d0) / (pi*pi)

      return
      end


      subroutine vp_jcp(xx,n, dvdx)
c derivatives of "cos + parabola"  - 2D function of x and y, periodic in x
      implicit real*8(a-h,o-z)
      dimension xx(n),dvdx(n)
      common/potcom1/d1,d2,d3,d4,d5,pi,twopi,twopi2
      
cc      if(n.ne.2) stop 'need n=2 in subroutine vp'

      x=xx(1)
      y=xx(2)

      dvdx(1)=-twopi*sin(twopi*x)*(1.0d0+d1*y)
     x     - twopi/d4*d3*sin(twopi*x/d4) + d5
      dvdx(2)=cos(twopi*x)*(d1) + d2*twopi2*y

      if(n.gt.2) then
        do i=3,n
           dvdx(1)=dvdx(1)-twopi*sin(twopi*x)*(1.0d0+d1*xx(i))
           dvdx(i)=cos(twopi*x)*d1 + d2*twopi2*xx(i)
        end do
      end if

      return
      end


      real*8 function vbias_jcp(xx, n, energy, dvdx, zmax, a)
c 2-D bias potential Eq. (31) of JCP paper
c take unbiased position and energy, return bias potential and 
c   derivative of V + V_bias
      implicit real*8(a-h,o-z)
      dimension xx(n),dvdx(n)
      common/potcom1/d1,d2,d3,d4,d5,pi,twopi,twopi2

c      if(n.ne.2) stop 'need n=2 in subroutine vbias_jcp'

      call vp_jcp(xx, n, dvdx)    ! unbiased dV/dx

      x=xx(1)
      y=xx(2)

c set up 2x2 Hessian and compute lowest eigenvalue
      h11 = -twopi2 * cos(twopi*x) * (1.d0+d1*y) -
     &     d3 * twopi2 * cos(twopi*x/d4) / (d4*d4)
      h12 = -twopi * d1 * sin(twopi*x)
      h22 = d2 * twopi2

      b = -h11 - h22
      c = h11 * h22 - h12 * h12
      eig1  = -0.5d0 * (b + sqrt(b*b - 4.d0*c))

      if (eig1 .gt. 0.d0) then     ! nonzero bias
         z = a * eig1 * eig1
         vbias_jcp = z / (1.d0 + z / zmax)

c     eigenvector C1 and Hessian derivatives (Eq. (10) of JCP paper)
         if (abs(h12) .lt. 1.d-12) then   ! special case: diagonal Hessian
            if (h11 .lt. h22) then
               c1x = 1.d0
               c1y = 0.d0
            else
               c1x = 0.d0
               c1y = 1.d0
            end if
         else
            c1y = sqrt(1.d0 / (1.d0 + (h12 / (eig1 - h11))**2))
            c1x = c1y * h12 / (eig1 - h11)
         end if
         dh11dx = twopi2 * twopi * sin(twopi*x) * (1.d0+d1*y) +
     &        d3 * ((twopi/d4)**3) * sin(twopi*x/d4)
         dh12dx = -twopi2 * d1 * cos(twopi*x)
         dh11dy = dh12dx

         dvdz = eig1 * (1.d0 + z / zmax)**-2
         dvdx(1) = dvdx(1) + 2.d0 * a * dvdz * c1x *
     &        (c1x * dh11dx + 2.d0 * c1y * dh12dx)
         dvdx(2) = dvdx(2) + 2.d0 * a * dvdz * c1x * c1x * dh11dy
      else
         vbias_jcp = 0.d0
      end if

      return
      end



      integer function kstate_jcp(x,n)
      implicit real*8(a-h,o-z)
      dimension x(n)

      kstate_jcp = nint(x(1)+0.5d0) 

      return
      end
     

c--------routines for 1-D Frenkel-Kontorova model---------

      subroutine potset_fk(ndim)
c  Set the parameters that define the potential.
      implicit real*8(a-h,o-z)
      common/potcom2/forcecon,pi,twopi,twopi2


      write(6,*) 'Enter the desired FK force constant:'
      read(5,*) forcecon

      write(6,*) '-------------------------------------------------'
      write(6,*) '-------------------------------------------------'
      write(6,*) 'icase=2: Frenkel-Kontorova model'
      write(6,*) 'vibrational period should be unity'
      write(6,*) '-------------------------------------------------'
      write(6,*) '-------------------------------------------------'

c some constants
      pi=2.0d0*acos(0.0d0)
      twopi=2.0d0*pi
      twopi2=twopi**2
  
      return
      end


      subroutine v_fk(x,n, energy)
      implicit real*8(a-h,o-z)
      dimension x(n)
      common/potcom2/forcecon,pi,twopi,twopi2

      energy=0.0d0
      do i=1,n
         energy = energy + (sin(pi*x(i)))**2
         if(i.lt.n) energy=energy+0.5d0*forcecon*(x(i+1)-x(i)-1.0d0)**2
      end do

      return
      end


      subroutine vp_fk(x,n, dvdx)
      implicit real*8(a-h,o-z)
      dimension x(n),dvdx(n)
      common/potcom2/forcecon,pi,twopi,twopi2

      do i=1,n
         dvdx(i) = 2.0d0*pi*(sin(pi*x(i)))*(cos(pi*x(i)))
         if(i.lt.n) dvdx(i)=dvdx(i) - forcecon*(x(i+1)-x(i)-1.0d0)
         if(i.gt.1) dvdx(i)=dvdx(i) + forcecon*(x(i)-x(i-1)-1.0d0)
      end do

      return
      end


      integer function kstate_fk(x,n)
      implicit real*8(a-h,o-z)
      dimension x(n)

      xave=0.0d0
      do i=1,n
         xave=xave+x(i)
      end do
      xave=xave/float(n)

      if(mod(n,2).eq.1) then
        kstate_fk = nint(xave) 
      else
        kstate_fk = nint(xave+0.5d0) 
      end if

      return
      end
